@@ -1,6 +1,5 @@
 /*
  * ndisc.c - ICMPv6 neighbour discovery command line tool
- * $Id: ndisc.c 625 2008-07-11 17:10:23Z remi $
  */
 
 /*************************************************************************
@@ -51,6 +50,11 @@
 #include <arpa/inet.h> /* inet_ntop() */
 #include <net/if.h> /* if_nametoindex() */
 
+#ifndef __linux__
+# include <net/if_dl.h> /* Link-Level sockaddr structure sockaddr_dl */
+# include <ifaddrs.h> /* getifaddrs and freeifaddrs*/
+#endif
+
 #include <netinet/in.h>
 #include <netinet/icmp6.h>
 
@@ -59,6 +63,13 @@
 # define IPV6_RECVHOPLIMIT IPV6_HOPLIMIT
 #endif
 
+/* BSD-like systems define ND_RA_FLAG_HA instead of ND_RA_FLAG_HOME_AGENT */
+#ifndef ND_RA_FLAG_HOME_AGENT
+# ifdef ND_RA_FLAG_HA
+#  define ND_RA_FLAG_HOME_AGENT ND_RA_FLAG_HA
+# endif
+#endif
+
 #ifndef AI_IDN
 # define AI_IDN 0
 #endif
@@ -116,6 +127,22 @@ sethoplimit (int fd, int value)
 	                    &value, sizeof (value))) ? -1 : 0;
 }
 
+static int
+setsourceip (int fd, const char *src, const char *ifname, int flags)
+{
+	struct sockaddr_in6 addr;
+
+	if (getipv6byname (src, ifname, (flags & NDISC_NUMERIC) ? 1 : 0, &addr))
+		return -1;
+
+	if (bind (fd, (const struct sockaddr *)&addr, sizeof (addr)))
+	{
+		perror (src);
+		return -1;
+	}
+	return 0;
+}
+
 
 static void
 printmacaddress (const uint8_t *ptr, size_t len)
@@ -163,6 +190,20 @@ getmacaddress (const char *ifname, uint8_t *addr)
 	memcpy (addr, req.ifr_hwaddr.sa_data, 6);
 	return 0;
 # else
+	/* No SIOCGIFHWADDR, which seems Linux specific. */
+	struct ifaddrs *ifa = NULL, *ifp;
+	getifaddrs(&ifa);
+	ifp = ifa;    /* preserve the address of ifa to free memory later */
+	while (ifp != NULL) {
+		if (ifp->ifa_addr->sa_family == AF_LINK && strcmp(ifp->ifa_name, ifname) == 0) {
+			const struct sockaddr_dl* sdl = (const struct sockaddr_dl*) ifp->ifa_addr;
+			memcpy(addr, sdl->sdl_data + sdl->sdl_nlen, 6);
+			freeifaddrs(ifa);
+			return 0;
+		}
+		ifp = ifp->ifa_next;
+	}
+	freeifaddrs(ifa);
 	(void)ifname;
 	(void)addr;
 	return -1;
@@ -270,7 +311,7 @@ static const unsigned ndisc_default = NDISC_VERBOSE1;
 static const char ndisc_usage[] = N_(
 	"Usage: %s [options] [IPv6 address] <interface>\n"
 	"Solicits on-link IPv6 routers (Router Discovery)\n");
-static const char ndisc_dataname[] = N_("advertised prefixes");
+static const char ndisc_dataname[] = N_("advertized prefixes");
 
 typedef struct nd_router_solicit solicit_packet;
 
@@ -321,6 +362,13 @@ parseprefix (const struct nd_opt_prefix_info *pi, size_t optlen, bool verbose)
 
 	if (verbose)
 	{
+		uint8_t v = pi->nd_opt_pi_flags_reserved;
+
+		printf (_("  On-link                 :          %3s\n"),
+		        gettext ((v & ND_OPT_PI_FLAG_ONLINK) ? N_ ("Yes") : N_("No")));
+		printf (_("  Autonomous address conf.:          %3s\n"),
+		        gettext ((v & ND_OPT_PI_FLAG_AUTO) ? N_ ("Yes") : N_("No")));
+
 		fputs (_("  Valid time              : "), stdout);
 		print32time (pi->nd_opt_pi_valid_time);
 		fputs (_("  Pref. time              : "), stdout);
@@ -398,6 +446,56 @@ parserdnss (const uint8_t *opt)
 }
 
 
+static int
+parsednssl (const uint8_t *opt)
+{
+	const uint8_t *base;
+	uint8_t optlen = opt[1];
+	if (optlen < 2)
+		return -1;
+
+	printf (_(" DNS search list          : "));
+
+	optlen *= 8;
+	optlen -= 8;
+	base = opt + 8;
+
+	for (unsigned i = 0; i < optlen; i++)
+	{
+		char str[256];
+
+		if (!base[i])
+			break;
+
+		do
+		{
+			if (base[i] + i + 1 >= optlen)
+			{
+				printf("\n");
+				return -1;
+			}
+
+			memcpy (str, &base[i + 1], base[i]);
+			str[base[i]] = 0;
+
+			i += base[i] + 1;
+
+			printf ("%s%s", str, base[i] ? "." : "");
+
+		} while (base[i]);
+
+		printf (" ");
+
+	}
+
+	printf("\n");
+
+	fputs (_("  DNS search list lifetime: "), stdout);
+	print32time (((const uint32_t *)opt)[1]);
+	return 0;
+}
+
+
 static int
 parseadv (const uint8_t *buf, size_t len, const struct sockaddr_in6 *tgt,
           bool verbose)
@@ -433,7 +531,11 @@ parseadv (const uint8_t *buf, size_t len, const struct sockaddr_in6 *tgt,
 		        gettext ((v & ND_RA_FLAG_MANAGED) ? N_ ("Yes") : N_("No")));
 		printf (_("Stateful other conf.      :          %3s\n"),
 		        gettext ((v & ND_RA_FLAG_OTHER) ? N_ ("Yes") : N_("No")));
+		printf (_("Mobile home agent         :          %3s\n"),
+		        gettext ((v & ND_RA_FLAG_HOME_AGENT) ? N_ ("Yes") : N_("No")));
 		printf (_("Router preference         :       %6s\n"), pref_i2n (v));
+		printf (_("Neighbor discovery proxy  :          %3s\n"),
+		        gettext ((v & 0x04) ? N_ ("Yes") : N_("No")));
 
 		/* Router lifetime */
 		fputs (_("Router lifetime           : "), stdout);
@@ -506,6 +608,10 @@ parseadv (const uint8_t *buf, size_t len, const struct sockaddr_in6 *tgt,
 			case 25: // RFC5006
 				parserdnss (ptr);
 				break;
+
+			case 31: // RFC6106
+				parsednssl (ptr);
+				break;
 		}
 		/* skips unrecognized option */
 
@@ -566,24 +672,27 @@ static ssize_t
 recvadv (int fd, const struct sockaddr_in6 *tgt, unsigned wait_ms,
          unsigned flags)
 {
-	struct timespec now, end;
+	struct timespec end;
 	unsigned responses = 0;
 
 	/* computes deadline time */
-	mono_gettime (&now);
+	mono_gettime (&end);
 	{
 		div_t d;
 		
 		d = div (wait_ms, 1000);
-		end.tv_sec = now.tv_sec + d.quot;
-		end.tv_nsec = now.tv_nsec + (d.rem * 1000000);
+		end.tv_sec += d.quot;
+		end.tv_nsec += d.rem * 1000000;
 	}
 
 	/* receive loop */
 	for (;;)
 	{
 		/* waits for reply until deadline */
+		struct timespec now;
 		ssize_t val = 0;
+
+		mono_gettime (&now);
 		if (end.tv_sec >= now.tv_sec)
 		{
 			val = (end.tv_sec - now.tv_sec) * 1000
@@ -601,10 +710,14 @@ recvadv (int fd, const struct sockaddr_in6 *tgt, unsigned wait_ms,
 
 		/* receives an ICMPv6 packet */
 		// TODO: use interface MTU as buffer size
-		uint8_t buf[1460];
+		union
+		{
+			uint8_t  b[1460];
+			uint64_t align;
+		} buf;
 		struct sockaddr_in6 addr;
 
-		val = recvfromLL (fd, buf, sizeof (buf), MSG_DONTWAIT, &addr);
+		val = recvfromLL (fd, &buf, sizeof (buf), MSG_DONTWAIT, &addr);
 		if (val == -1)
 		{
 			if (errno != EAGAIN)
@@ -617,7 +730,7 @@ recvadv (int fd, const struct sockaddr_in6 *tgt, unsigned wait_ms,
 		 && (addr.sin6_scope_id != tgt->sin6_scope_id))
 			continue;
 
-		if (parseadv (buf, val, tgt, (flags & NDISC_VERBOSE) != 0) == 0)
+		if (parseadv (buf.b, val, tgt, (flags & NDISC_VERBOSE) != 0) == 0)
 		{
 			if (flags & NDISC_VERBOSE)
 			{
@@ -634,7 +747,6 @@ recvadv (int fd, const struct sockaddr_in6 *tgt, unsigned wait_ms,
 			if (flags & NDISC_SINGLE)
 				return 1 /* = responses */;
 		}
-		mono_gettime (&now);
 	}
 
 	return -1; /* error */
@@ -645,7 +757,7 @@ static int fd;
 
 static int
 ndisc (const char *name, const char *ifname, unsigned flags, unsigned retry,
-       unsigned wait_ms)
+       unsigned wait_ms, const char *source)
 {
 	struct sockaddr_in6 tgt;
 
@@ -673,6 +785,10 @@ ndisc (const char *name, const char *ifname, unsigned flags, unsigned retry,
 	setsockopt (fd, IPPROTO_IPV6, IPV6_RECVHOPLIMIT,
 	            &(int){ 1 }, sizeof (int));
 
+	/* sets source address */
+	if ((source != NULL) && setsourceip (fd, source, ifname, flags))
+		goto error;
+
 	/* resolves target's IPv6 address */
 	if (getipv6byname (name, ifname, (flags & NDISC_NUMERIC) ? 1 : 0, &tgt))
 		goto error;
@@ -756,6 +872,7 @@ usage (const char *path)
 "  -n, --numeric  don't resolve host names\n"
 "  -q, --quiet    only print the %s (mainly for scripts)\n"
 "  -r, --retry    maximum number of attempts (default: 3)\n"
+"  -s, --source   specify source IPv6 address\n"
 "  -V, --version  display program version and exit\n"
 "  -v, --verbose  verbose display (this is the default)\n"
 "  -w, --wait     how long to wait for a response [ms] (default: 1000)\n"
@@ -769,7 +886,7 @@ static int
 version (void)
 {
 	printf (_(
-"ndisc6: IPv6 Neighbor/Router Discovery userland tool %s (%s)\n"), VERSION, "$Rev: 625 $");
+"ndisc6: IPv6 Neighbor/Router Discovery userland tool %s (%s)\n"), VERSION, "$Rev$");
 	printf (_(" built %s on %s\n"), __DATE__, PACKAGE_BUILD_HOSTNAME);
 
 	printf (_("Configured with: %s\n"), PACKAGE_CONFIGURE_INVOCATION);
@@ -791,6 +908,7 @@ static const struct option opts[] =
 	{ "numeric",  no_argument,       NULL, 'n' },
 	{ "quiet",    no_argument,       NULL, 'q' },
 	{ "retry",    required_argument, NULL, 'r' },
+	{ "source",   required_argument, NULL, 's' },
 	{ "version",  no_argument,       NULL, 'V' },
 	{ "verbose",  no_argument,       NULL, 'v' },
 	{ "wait",     required_argument, NULL, 'w' },
@@ -815,9 +933,9 @@ main (int argc, char *argv[])
 
 	int val;
 	unsigned retry = 3, flags = ndisc_default, wait_ms = nd_delay_ms;
-	const char *hostname, *ifname;
+	const char *hostname, *ifname, *source = NULL;
 
-	while ((val = getopt_long (argc, argv, "1hmnqr:Vvw:", opts, NULL)) != EOF)
+	while ((val = getopt_long (argc, argv, "1hmnqr:s:Vvw:", opts, NULL)) != EOF)
 	{
 		switch (val)
 		{
@@ -851,6 +969,10 @@ main (int argc, char *argv[])
 				retry = l;
 				break;
 			}
+
+			case 's':
+				source = optarg;
+				break;
 				
 			case 'V':
 				return version ();
@@ -903,6 +1025,6 @@ main (int argc, char *argv[])
 		return quick_usage (argv[0]);
 
 	errno = errval; /* restore socket() error value */
-	return -ndisc (hostname, ifname, flags, retry, wait_ms);
+	return -ndisc (hostname, ifname, flags, retry, wait_ms, source);
 }
 
