@@ -1,6 +1,5 @@
 /*
  * traceroute.c - TCP/IPv6 traceroute tool
- * $Id: traceroute.c 624 2008-05-01 14:26:09Z remi $
  */
 
 /*************************************************************************
@@ -176,7 +175,7 @@ recv_payload (int fd, void *buf, size_t len,
 	     cmsg = CMSG_NXTHDR (&hdr, cmsg))
 		if ((cmsg->cmsg_level == IPPROTO_IPV6)
 		 && (cmsg->cmsg_type == IPV6_HOPLIMIT))
-			memcpy (hlim, CMSG_DATA (cmsg), sizeof (hlim));
+			memcpy (hlim, CMSG_DATA (cmsg), sizeof (*hlim));
 
 	return val;
 }
@@ -1008,7 +1007,7 @@ traceroute (const char *dsthost, const char *dstport,
 	memset (&dst, 0, sizeof (dst));
 	if (connect_proto (protofd, &dst, dsthost, dstport, srchost, srcport))
 		goto error;
-	printf (_("%u hops max, "), max_ttl);
+	printf (ngettext ("%u hop max, ", "%u hops max, ", max_ttl), max_ttl);
 
 #ifdef SO_ATTACH_FILTER
 	{
@@ -1059,7 +1058,8 @@ traceroute (const char *dsthost, const char *dstport,
 	if (packet_len < overhead)
 		packet_len = overhead;
 
-	printf (_("%lu byte packets\n"), (unsigned long)packet_len);
+	printf (ngettext ("%zu byte packets\n", "%zu bytes packets\n", packet_len),
+            packet_len);
 	packet_len -= overhead;
 
 	struct timespec delay_ts;
@@ -1249,7 +1249,7 @@ static int
 version (void)
 {
 	printf (_(
-"traceroute6: TCP & UDP IPv6 traceroute tool %s (%s)\n"), VERSION, "$Rev: 624 $");
+"traceroute6: TCP & UDP IPv6 traceroute tool %s (%s)\n"), VERSION, "$Rev$");
 	printf (_(" built %s on %s\n"), __DATE__, PACKAGE_BUILD_HOSTNAME);
 	printf (_("Configured with: %s\n"), PACKAGE_CONFIGURE_INVOCATION);
 	puts (_("Written by Remi Denis-Courmont\n"));
